/* graph6io.c
 *
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "graph_config.h"

/**
 * @file
 * @brief Provides input/output routines translating to and from graph6 format
 */

/**
 * Reads a single graph from the given file and stores it as a list of neighbours
 * @param[in]  f        file pointer
 * @param[out] graph    list of neighbours of all vertices, 0-based
 * @param[out] degrees  degrees of the vertices
 * @param[out] nv       number of vertices in te graph
 * @return 0 if end of file, 1 if OK, 2 if in error
 */
int read_graph6_adj_list(FILE *f, ADJ_LIST_PTR graph, unsigned int *degrees, int *nv) {
    int n;
    char c[3];
    int byte;
    do { /* skip newline(s) */
        if((byte = fgetc(f)) == EOF) {
            return (feof(f) ? 0 : 2);
        }
    } while(byte < 63 || byte > 126);
    if(byte == 126) {
        if(fread(&c[0], sizeof (char), 3, f) < 3) {
            return (2);
        }
        n = ((unsigned int) c[0] - 63)*4096 + ((unsigned int) c[1] - 63)*64 + ((unsigned int) c[2] - 63);
    } else {
        n = byte - 63;
    }

    if(n > MAXN) {
        fprintf(stderr, "Error: too many vertices: %d\n", n);
        exit(1);
    }
    *nv = n;

    unsigned int i, j;
    for(i = 0; i < n; i++) {
        degrees[i] = 0;
    }
    byte = 0;

    unsigned int k = 0;
    for(j = 1; j < n; j++) {
        for(i = 0; i < j; i++) {
            if(k == 0) {
                if((byte = fgetc(f)) == EOF) {
                    return (2);
                }
                byte -= 63;
                k = 32;
            }
            if(byte & k) {
                graph[i][degrees[i]] = j;
                graph[j][degrees[j]] = i;
                degrees[i]++;
                degrees[j]++;
            }
            k = k >> 1;
        }
    }
    return (1);
}

//Important: must be ints, not chars since nv can be > 254
static
int are_neighbours(ADJ_LIST_PTR current_graph, unsigned int* degrees, int v1, int v2) {
    int i;
    for(i = 0; i < degrees[v1]; i++)
        if(current_graph[v1][i] == v2)
            return 1;

    return 0;
}

/**
 * Outputs a single graph which is represented as a list of neighbours in graph6 format to the given file
 * @param[in]  f        file pointer
 * @param[in] graph    list of neighbours of all vertices, 0-based
 * @param[in] degrees  degrees of the vertices
 * @param[in] nv       number of vertices in te graph
 * @return 0 if OK, 1 if error
 */
int write_graph6_adj_list(FILE *f, ADJ_LIST_PTR graph, unsigned int *degrees, int nv) {
    int i, j, k = 32;
    char byte = 0;
    if(nv <= 62) {
        fprintf(f, "%c", nv + 63);
    }/* N(n) */
    else {
        fprintf(f, "%c%c%c%c", 126, nv / 4096 + 63, (nv % 4096) / 64 + 63, nv % 64 + 63);
    } /* N(n) */
    for(j = 1; j < nv; j++) {
        for(i = 0; i < j; i++) {
            // TODO: make this more efficient
            if(are_neighbours(graph, degrees, i, j)) {
                byte |= k;
            }
            k = k >> 1;
            if(k == 0) {
                fprintf(f, "%c", byte + 63);
                k = 32;
                byte = 0;
            }
        }
    }
    if(k != 32) {
        fprintf(f, "%c", byte + 63);
    }
    fprintf(f, "\n");
    return (ferror(f) ? 0 : 1);
}

