/* longmod.c
 *
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file
 * Computes ranks/inverses of adjacency matrices modulo a large 'long' prime
 */

#include <memory.h>
#include <stdio.h>
#include "graph_config.h"

#define LONG_PRIME 4294967291  // or 2147483647 = 2^31-1
//#define LONG_PRIME 2147483647  // = 2^31-1

static INT_ADJ_MATRIX inverse;

static unsigned int inverse_sum[MAXN]; // *minus* the sum of all rows of inverse

/**
 * Computes the inverse of a modulo PRIME
 */
static
unsigned int invert(unsigned int a) {
    unsigned long b = 1;
    while (a > 1) {
        unsigned int f = LONG_PRIME / a + 1;
        b = b * f % LONG_PRIME;
        a = a - LONG_PRIME % a;
    }
    return (unsigned int) b;
}

/*
 * Rescales a row
 */
static
void row_scale(unsigned int *row, int nv, unsigned int factor) {
    for (int i = 0; i < nv; i++) {
        row[i] = (unsigned long) row[i] * factor % LONG_PRIME;
    }
}

/*
 * Reduces destination row
 */
static
void row_reduce(int nv, INT_ADJ_MATRIX_PTR matrix, int dest, int src, unsigned int src_factor) {
    if (src_factor != 0) {
        for (int i = 0; i < nv; i++) {
            matrix[dest][i] =
                    (matrix[dest][i] + ((unsigned long) LONG_PRIME - matrix[src][i]) * src_factor) % LONG_PRIME;
        }
    }
}

// for debugging
static
void print_adj_matrix(FILE *f, int nv, INT_ADJ_MATRIX_PTR matrix) {
    for (int i = 0; i < nv; i++) {
        for (int j = 0; j < nv; j++)
            fprintf(f, "%d ", matrix[i][j]);
        fprintf(f, "\n");
    }
}

/**
 * Interchanges two rows
 */
static
void row_interchange(int nv, INT_ADJ_MATRIX_PTR matrix, int r1, int r2) {
    for (int i = 0; i < nv; i++) {
        unsigned int tmp = matrix[r1][i];
        matrix[r1][i] = matrix[r2][i];
        matrix[r2][i] = tmp;
    }
}

static
int inverse_longmod_matrix_aux(int nv, INT_ADJ_MATRIX_PTR matrix, INT_ADJ_MATRIX_PTR result) {
    // start with the identity matrix
    memset(result, 0, MAXN * MAXN * sizeof(int));
    for (int i = 0; i < nv; i++) {
        result[i][i] = 1;
    }

    // do the reduction. Apply all elementary operations to both matrix and inverse
    for (int column = 0; column < nv; column++) {
        // find pivot
        int k = column;
        while (k < nv && matrix[k][column] == 0)
            k++;
        if (k < nv) {
            if (k > column) {
                row_interchange(nv, matrix, k, column);
                row_interchange(nv, result, k, column);
            }
            if (matrix[column][column] != 1) {
                unsigned int factor = invert(matrix[column][column]);
                row_scale(matrix[column], nv, factor);
                // important! do not inline factor, for matrix[column][column] becomes 1 after the above call
                row_scale(result[column], nv, factor);
            }
            for (int dest = 0; dest < nv; dest++) {
                if (dest != column) {
                    int src_factor = matrix[dest][column];
                    row_reduce(nv, matrix, dest, column, src_factor);
                    // important! do not inline src_factor, for matrix[dest][column] becomes 0 after the above call
                    row_reduce(nv, result, dest, column, src_factor);
                }
            }
        } else {
            return 0; // singular
        }

    }
    return 1; // true
}


/**
 * Compute the inverse of the given matrix and stores it locally for use in other methods
 * @param[in out] matrix matrix to be inverted
 * @return 1 when matrix is nonsingular, 0 otherwise
 */
int inverse_longmod_matrix(int nv, INT_ADJ_MATRIX_PTR matrix) {
    if (inverse_longmod_matrix_aux(nv, matrix, inverse)) {
        // precompute minus the sum of all rows of inverse - for use in longmod_norm_bit_vector
        for (int j = 0; j < nv; j++) {
            inverse_sum[j] = inverse[0][j];
        }
        for (int i = 1; i < nv; i++) {
            for (int j = 0; j < nv; j++) {
                inverse_sum[j] = ((long) inverse_sum[j] + inverse[i][j]) % LONG_PRIME;
            }
        }
        for (int j = 0; j < nv; j++) {
            if (inverse_sum[j]) {
                inverse_sum[j] = LONG_PRIME - inverse_sum[j];
            }
        }

        return 1;
    } else {
        return 0;
    }
}

int longmod_is_nucifer(int nv, INT_ADJ_MATRIX_PTR matrix) {
    INT_ADJ_MATRIX result;
    if (inverse_longmod_matrix_aux(nv, matrix, result)) {
        // check diagonal
        for (int i = 0; i < nv; ++i) {
            if (result[i][i] != 0) {
                return 0;
            }
        }
        // check others (matrix is symmetric)
        for (int i = 1; i < nv; ++i) {
            for (int j = 0; j < i; ++j) {
                if (result[i][j] == 0) {
                    return 0;
                }
            }
        }
        return 1;
    } else {
        return 0;
    }

}

/**
 * Computes bMb^T for a bit vector b and a matrix M.
 */
unsigned long longmod_norm_bit_vector(unsigned int b, int nv) {
    unsigned long norm = 0;
    for (int i = 0, bitsi = b; i < nv; i++, bitsi <<= 1) {
        if ((bitsi & 0x80000000) != 0) {
            for (int j = 0, bitsj = b; j < nv; j++, bitsj <<= 1) {
                if ((bitsj & 0x80000000) != 0) {
                    norm = (norm + inverse[i][j]) % LONG_PRIME;
                }
            }
        }
    }
    return norm;
}

/**
 * Checks for nut property: bM cannot contain a 0 and bMb^T must be zero.
 * @return whether the nut property holds
 */
int longmod_nut_property(unsigned int b, int nv) {

    // note: uses -inverse as M. Optimized of the case where
    // b contains more zeroes than ones. Uses precomputed inverse_sum

    // compute bM
    unsigned int vector[MAXN];
    memcpy(vector, inverse_sum, nv * sizeof(unsigned int));
    for (int i = 0, bitsi = b; i < nv; i++, bitsi <<= 1) {
        if ((bitsi & 0x80000000) == 0) {
            for (int j = 0; j < nv; j++) {
                vector[j] = ((unsigned long) vector[j] + inverse[i][j]) % LONG_PRIME;
            }
        }
    }
    // check for zeroes
    for (int j = 0; j < nv; j++) {
        if (vector[j] == 0) {
            return 0; // not a nut
        }
    }
    // compute norm
    unsigned long norm = 0;
    for (int i = 0, bitsi = b; i < nv; i++, bitsi <<= 1) {
        if ((bitsi & 0x80000000) != 0) {
            norm = (norm + vector[i]) % LONG_PRIME;
        }
    }
    return norm == 0;
}

static unsigned long long FREQS[MAXN];

/**
 * Same as @ref{#longmod_nut_property} but also counts the degree distribution  of the
 * vertices added by geng in the last generation step. Used for profiling
 */
int longmod_nut_property_with_count(unsigned int b, int nv) {
    int count = 0;
    for (int i = 0, bitsi = b; i < nv; i++, bitsi <<= 1) {
        if ((bitsi & 0x80000000) != 0) {
            count++;
        }
    }
    FREQS[count]++;
    return longmod_nut_property(b, nv);
}

/**
 * Output the result of counting with @ret{#longmod_nut_property_with_count}
 */
void longmod_print_counts(int nv) {
    fprintf(stderr, "Last step degree distribution:\n");
    for (int i = 0; i <= nv; i++) {
        fprintf(stderr, "  deg %2d: %llu\n", i, FREQS[i]);
    }
}


/*
 * Reduces destination row
 */
static
void row_reduce_dest_factor(int nv, INT_ADJ_MATRIX_PTR matrix, int dest, unsigned int dest_factor, 
        int src, unsigned int src_factor) {
    if (src_factor != 0) {
        for (int i = 0; i < nv; i++) {
            matrix[dest][i] = ((unsigned long) matrix[dest][i] * dest_factor) % LONG_PRIME; //Extra
            matrix[dest][i] = 
                    (matrix[dest][i] + ((unsigned long) LONG_PRIME - matrix[src][i]) * src_factor) % LONG_PRIME;
            //mpz_mul(matrix[dest][i], matrix[dest][i], dest_factor);
            //mpz_submul(matrix[dest][i], matrix[src][i], src_factor);
        }
    }
}

/**
 * Performs a reduction of the given matrix and returns the rank
 */
int longmod_reduce_matrix(int nv, INT_ADJ_MATRIX_PTR matrix) {
    int row = 0;
    for (int column = 0; column < nv; column++) {
        // find pivot
        int k = row;
        while (k < nv && matrix[k][column] == 0)
            k++;
        if (k < nv) {
            if (k > row) {
                row_interchange(nv, matrix, k, row);
            }
            for (int dest = row + 1; dest < nv; dest++) {
                row_reduce_dest_factor(nv, matrix, dest, matrix[row][column], row, matrix[dest][column]);
                //row_reduce(nv, matrix, dest, row, matrix[dest][column]);
            }
            row++;
        }
    }
    return row;
}



/**
 * Check whether the given matrix has a kernel vector without zeroes. Uses
 * backsubstitution.
 * @param[in] nv number of rows and columns in the matrix
 * @param[in] matrix in echelon form, must be of nullity 1
 * @return whether the matrix has a kernel vector without zeroes
 */
int check_zeroless_kernel_vector_longmod(int nv, INT_ADJ_MATRIX_PTR matrix) {

    // initial check: 'diagonal' must be nonzero - only last element need be checked
    if (matrix[nv - 2][nv - 2] == 0) {
        return 0;
    }

    unsigned int vector[nv];
    for(int i = 0; i < nv; i++)
        vector[i] = 0;

    vector[nv - 1] = 1;

    unsigned int sum;
    for (int i = nv - 2; i >= 0; i--) {
        sum = 0;
        for (int j = i + 1; j < nv; j++) {
            sum = (sum + ((unsigned long) LONG_PRIME - vector[j]) * matrix[i][j]) % LONG_PRIME;
            vector[j] = ((unsigned long) vector[j] * matrix[i][i]) % LONG_PRIME;
            //mpz_submul(sum, vector[j], matrix[i][j]);
            //mpz_mul(vector[j], vector[j], matrix[i][i]);            
        }
        if (sum == 0) {
            return 0; // zero found
        }
        vector[i] = sum;
    }
    
    return 1; // no zero found
}
