/* modulo.c
 *
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file
 * Computes ranks/inverses of adjacency matrices modulo a large prime
 */

#include <memory.h>
#include <stdio.h>
#include "graph_config.h"

#define PRIME 65521
#define PRIMITIVE_ELEMENT 17
//#define PRIME 5
//#define PRIMITIVE_ELEMENT 2

unsigned int INVERSES[PRIME];

/**
 * Initializes some tables.
 * @return whether initialization succeeded
 */
int init_modulo () {
    unsigned int prod = PRIMITIVE_ELEMENT;
    unsigned int lastprod = 1;
    for (unsigned int i=2; i < PRIME; i++) {
        lastprod = prod;
        //fprintf (stderr, "%d\n", prod);
        prod = prod * PRIMITIVE_ELEMENT % PRIME;
        if (lastprod == 1) {
            return 0; // not a primitive element
        }
    }

    // lastprod is now the inverse of the primitive element
    unsigned int p = 1;
    unsigned int q = 1;
    for (unsigned int i=1; i < PRIME; i++) {
//        if ( p*q % PRIME != 1) {
//            fprintf(stderr, "Not inverses: %d, %d\n", p, q);
//        }
        INVERSES[p] = q;
        p = p * PRIMITIVE_ELEMENT % PRIME;
        q = q * lastprod % PRIME;
    }

//    fprintf(stderr, "Inverses: ");
//    for (int i = 0; i < PRIME; ++i) {
//        fprintf (stderr, "%d ", INVERSES[i]);
//    }
//    fprintf(stderr, "\n");
    return 1; // all is OK
}


static INT_ADJ_MATRIX inverse;


/*
 * Rescales a row
 */
static
void row_scale(unsigned int *row, int nv, int factor) {
    for (int i = 0; i < nv; i++) {
        row[i] = row[i] * factor % PRIME;
    }
}

/*
 * Reduces destination row
 */
static
void row_reduce(int nv, INT_ADJ_MATRIX_PTR matrix, int dest, int src, int src_factor) {
    if (src_factor != 0) {
        for (int i = 0; i < nv; i++) {
            matrix[dest][i] = (matrix[dest][i] + (PRIME - matrix[src][i]) * src_factor) % PRIME;
        }
    }
}

// for debugging
static
void print_adj_matrix(FILE *f, int nv, INT_ADJ_MATRIX_PTR matrix) {
    for (int i = 0; i < nv; i++) {
        for (int j = 0; j < nv; j++)
            fprintf(f, "%d ", matrix[i][j]);
        fprintf(f, "\n");
    }
}

/**
 * Interchanges two rows
 */
static
void row_interchange(int nv, INT_ADJ_MATRIX_PTR matrix, int r1, int r2) {
    for (int i = 0; i < nv; i++) {
        unsigned int tmp = matrix[r1][i];
        matrix[r1][i] = matrix[r2][i];
        matrix[r2][i] = tmp;
    }
}

/**
 * Performs a reduction of the given matrix and returns the rank
 */
int modulo_reduce_matrix(int nv, INT_ADJ_MATRIX_PTR matrix) {
    int row = 0;
    for (int column = 0; column < nv; column++) {
        // find pivot
        int k = row;
        while (k < nv && matrix[k][column] == 0)
            k++;
        if (k < nv) {
            if (k > row) {
                row_interchange(nv, matrix, k, row);
            }
            if (matrix[row][column] != 1) {
                row_scale(matrix[row], nv, INVERSES[matrix[row][column]]);
            }
            for (int dest = row + 1; dest < nv; dest++) {
                row_reduce(nv, matrix, dest, row, matrix[dest][column]);
            }
            row++;
        }
    }
    return row;
}

/**
 * Compute the inverse of the given matrix and stores it locally for use in other methods
 * @param[in out] matrix matrix to be inverted
 * @return 1 when matrix is nonsingular, 0 otherwise
 */
int inverse_modulo_matrix(int nv, INT_ADJ_MATRIX_PTR matrix) {

//    fprintf (stderr, "MATRIX = \n");
//    print_adj_matrix(stderr, nv, matrix);

    // start with the identity matrix
    memset(inverse, 0, MAXN * MAXN * sizeof(int));
    for (int i = 0; i < nv; i++) {
        inverse[i][i] = 1;
    }

    // do the reduction. Apply all elementary operations to both matrix and inverse
    for (int column = 0; column < nv; column++) {
        // find pivot
        int k = column;
        while (k < nv && matrix[k][column] == 0)
            k++;
        if (k < nv) {
            if (k > column) {
                row_interchange(nv, matrix, k, column);
                row_interchange(nv, inverse, k, column);
            }
            if (matrix[column][column] != 1) {
                unsigned int factor = INVERSES[matrix[column][column]];
                row_scale(matrix[column], nv, factor);
                // important! do not inline factor, for matrix[column][column] becomes 1 after the above call
                row_scale(inverse[column], nv, factor);
            }
            for (int dest = 0; dest < nv; dest++) {
                if (dest != column) {
                    int src_factor = matrix[dest][column];
                    row_reduce(nv, matrix, dest, column, src_factor);
                    // important! do not inline src_factor, for matrix[dest][column] becomes 0 after the above call
                    row_reduce(nv, inverse, dest, column, src_factor);
                }
            }
        } else {
            return 0; // singular
        }

    }
//    fprintf (stderr, "INVERSE = \n");
//    print_adj_matrix(stderr, nv, inverse);
    return 1; // true
}

/**
 * Computes bMb^T for a bit vector b and an n x n  matrix M.
 */
unsigned int modulo_norm_bit_vector(unsigned int b, int n) {
    unsigned int norm = 0;
    for (int i = 0, bitsi = b; i < n; i++, bitsi <<= 1) {
        if ((bitsi & 0x80000000) != 0) {
            for (int j = 0, bitsj = b; j < n; j++, bitsj <<= 1) {
                if ((bitsj & 0x80000000) != 0) {
                    norm = (norm + inverse[i][j]) % PRIME;
                }
            }
        }
    }
    return norm;
}

