/* mp.c
 *
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file
 * Multiprecision adjacency matrices.
 */

#include <stdio.h>
#include "mp.h"

/**
 * Convert a neighbour list to a multiprecision adjacency matrix with entries initialized.
 * Use @ref clear_mp_matrix when finished
 */
void to_mp_matrix(ADJ_LIST_PTR graph, unsigned int *degrees, int nv, MP_MATRIX_PTR matrix) {
    for (int i = 0; i < nv; i++) {
        for (int j = 0; j < nv; j++) {
            mpz_init(matrix[i][j]);
        }
    }

    for (int i = 0; i < nv; i++) {
        for (int j = 0; j < degrees[i]; j++) {
            mpz_set_ui(matrix[i][graph[i][j]], 1);
        }
    }
}

/**
 * Convert a binary graph representation to a multiprecision adjacency matrix with entries initialized.
 * Important! Matrix must already have been inited!
 */
void bits_to_old_mp_matrix(unsigned int *graph, int nv, MP_MATRIX_PTR matrix) {
    for (int i = 0; i < nv; i++) {
        for (int j = 0; j < nv; j++) {
            mpz_set_ui(matrix[i][j], 0);
        }
    }

    for (int i = 0; i < nv; i++) {
        for (int j = 0, bits = graph[i]; j < nv; j++, bits <<= 1) {
            if ((bits & 0x80000000) != 0) {
                mpz_set_ui(matrix[i][j], 1);
            }
        }
    }
}

/**
 * Initialize the full multiprecision adjacency matrix
 */
void init_mp_matrix (MP_MATRIX_PTR matrix) {
    for (int i = 0; i < MAXN; i++) {
        for (int j = 0; j < MAXN; j++) {
            mpz_init(matrix[i][j]);
        }
    }

}


/**
 * Cleans memory used by a multiprecision adjacency matrix. Call when the matrix contents
 * are no longer needed.
 */
void clear_mp_matrix(int nv, MP_MATRIX_PTR matrix) {
    for (int i = 0; i < nv; i++) {
        for (int j = 0; j < nv; j++) {
            mpz_clear(matrix[i][j]);
        }
    }
}

/**
 * Print a multiprecision matrix
 */
void print_mp_matrix(FILE *f, int nv, MP_MATRIX_PTR matrix) {
    for (int i = 0; i < nv; i++) {
        for (int j = 0; j < nv; j++) {
            gmp_fprintf(f, " %Zd", matrix[i][j]);
        }
        fprintf(f, "\n");
    }
    fprintf(f, "\n");
}

/*
 * Reduces destination row
 */
static
void row_reduce(int nv, MP_MATRIX_PTR matrix, int dest, mpz_t dest_factor, int src, mpz_t src_factor) {
    if (mpz_sgn(src_factor) != 0) {
        for (int i = 0; i < nv; i++) {
            mpz_mul(matrix[dest][i], matrix[dest][i], dest_factor);
            mpz_submul(matrix[dest][i], matrix[src][i], src_factor);
        }
    }
}

/**
 * Interchanges two rows
 */
static
void row_interchange(int nv, MP_MATRIX_PTR matrix, int r1, int r2) {
    for (int i = 0; i < nv; i++) {
        mpz_swap(matrix[r1][i], matrix[r2][i]);
    }
}

/**
 * Performs a reduction of the given matrix and returns the rank
 */
int reduce_mp_matrix(int nv, MP_MATRIX_PTR matrix) {
    mpz_t src_factor;
    mpz_init(src_factor);
    int row = 0;
    for (int column = 0; column < nv; column++) {
        // find pivot
        int k = row;
        while (k < nv && mpz_sgn(matrix[k][column]) == 0)
            k++;
        if (k < nv) {
            if (k > row) {
                row_interchange(nv, matrix, k, row);
            }
            for (int dest = row + 1; dest < nv; dest++) {
                mpz_set(src_factor, matrix[dest][column]);
                row_reduce(nv, matrix, dest, matrix[row][column], row, src_factor);
            }
            row++;
        }
    }
    mpz_clear(src_factor);
    return row;
}

static
void clear_mp_vector(int nv, mpz_t *vector) {
    for (int i = 0; i < nv; i++) {
        mpz_clear(vector[i]);
    }
}

static
void init_mp_vector(int nv, mpz_t *vector) {
    for (int i = 0; i < nv; i++) {
        mpz_init(vector[i]);
    }
}

/**
 * Check whether the given matrix has a kernel vector without zeroes. Uses
 * backsubstitution.
 * @param[in] nv number of rows and columns in the matrix
 * @param[in] matrix in echelon form, must be of nullity 1
 * @return whether the matrix has a kernel vector without zeroes
 */
int check_zeroless_kernel_vector_mp_matrix(int nv, MP_MATRIX_PTR matrix) {

    // initial check: 'diagonal' must be nonzero - only last element need be checked
    if (mpz_sgn(matrix[nv - 2][nv - 2]) == 0) {
        return 0;
    }

    // TODO: move init and clear (twice) outside the method but do not forget to zero the vector explicitely when reusing!
    mpz_t vector[nv];
    init_mp_vector(nv, vector);

    mpz_set_si(vector[nv - 1], 1);

    mpz_t sum;
    mpz_init(sum);
    for (int i = nv - 2; i >= 0; i--) {
        mpz_set_si(sum, 0);
        for (int j = i + 1; j < nv; j++) {
            mpz_submul(sum, vector[j], matrix[i][j]);
            mpz_mul(vector[j], vector[j], matrix[i][i]);
        }
        if (mpz_sgn(sum) == 0) {
            mpz_clear(sum);
            clear_mp_vector(nv, vector);
            return 0; // zero found
        }
        mpz_set(vector[i], sum);
    }
    mpz_clear(sum);
    clear_mp_vector(nv, vector);
    return 1; // no zero found
}

/**
 * Reduce the invertible matrix to diagonal form by left multiplication and compute the
 * corresponding multiplication matrix.
 * @param[in out] matrix matrix to be reduced
 * @param[out] result resulting multiplication matrix
 * @return whether the matrix was indeed invertible
 */
int diagonalize_old_mp_matrix(int nv, MP_MATRIX_PTR matrix, MP_MATRIX_PTR result) {

    // start with the identity matrix
    for (int i=0; i < nv; i++) {
        for (int j=0; j < nv; j++) {
            mpz_set_ui (result[i][j], 0);
        }
        mpz_set_ui (result[i][i], 1);
    }

    // do the reduction. Apply all elementary operations to both matrix and result
    mpz_t src_factor;
    mpz_init(src_factor);
    for (int column = 0; column < nv; column++) {
        // find pivot
        int k = column;
        while (k < nv && mpz_sgn(matrix[k][column]) == 0)
            k++;
        if (k < nv) {
            if (k > column) {
                row_interchange(nv, matrix, k, column);
                row_interchange(nv, result, k, column);
            }
            for (int dest = 0; dest < nv; dest++) {
                if (dest != column) {
                    mpz_set(src_factor, matrix[dest][column]);
                    row_reduce(nv, matrix, dest, matrix[column][column], column, src_factor);
                    row_reduce(nv, result, dest, matrix[column][column], column, src_factor);
                }
            }
        } else {
            return 0; // singular
        }
    }
    mpz_clear(src_factor);
    return 1; // true
}


