/* nucifergen.c
 *
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file
 * Program which generates nuciferous graphs only. Version of geng with
 * additional filters to only generate nut graphs.
 *
 * Usage example: ./nucifer -d2 -D3 -q -c 15 | wc -l
 */

#include <stdio.h>
#include "libs/pgeng.h"

#include "graph_config.h"
#include "graphutils.h"
#include "longmod.h"

static ADJ_MATRIX adj_matrix;

static ADJ_LIST adj_list;
static DEGREES degrees;    

int prune_print_graph(unsigned int *g, int n, int maxn) {
    if (n == maxn) {
        bits_to_adj_matrix(g, n, adj_matrix);
        print_adj_matrix(stderr, n, adj_matrix);
        fprintf(stderr, "\n");
    }
    return 0; // everything is OK
}

static INT_ADJ_MATRIX int_matrix;

static int pseudo_inverse_valid;

static unsigned long long counter[3];

static unsigned long long num_graphs_generated[MAXN+1];
static unsigned long long num_graphs_accepted[MAXN+1];

static unsigned long long int can_reject_nucifer_G_minus_last = 0;
static unsigned long long int cannot_reject_nucifer_G_minus_last = 0;

//Using preprune on last level is a bit faster
int preprune_singular_parent_tracked_longmod_nucifer(unsigned int *g, int n, int maxn) {
    if(n == maxn) {
        if (!longmod_nut_property(g[n - 1], n - 2)) {
            can_reject_nucifer_G_minus_last++;
            return 1; // G-{n-1} is not a nut
        } 
    } 
    
    return 0;
}

int prune_singular_parent_tracked_longmod_nucifer(unsigned int *g, int n, int maxn) {
    if(n == maxn) {
        //Was already done in preprune!
        /*
        if (!longmod_nut_property(g[n - 1], n - 2)) {
            can_reject_nucifer_G_minus_last++;
            return 1; // G-{n-1} is not a nut
        } else {
            cannot_reject_nucifer_G_minus_last++;
            bits_to_int_adj_matrix(g, n, int_matrix);
            return ! longmod_is_nucifer(n, int_matrix);
        }
        */
        cannot_reject_nucifer_G_minus_last++;
        return ! longmod_is_nucifer(n, int_matrix);
    } else if (n == maxn - 1) {
        if (longmod_nut_property(g[n - 1], n - 1)) {
            return 0; // this is a nut graph
        } else {
            return 1;
        }
    } else if (n == maxn - 2) {
        bits_to_int_adj_matrix(g, n, int_matrix);

        int ret_code = inverse_longmod_matrix(n, int_matrix);
        if (ret_code == 1) {
            pseudo_inverse_valid = 1;
            counter[0]++;
            return 0; // nonsingular
        } else {
            pseudo_inverse_valid = 0;
            counter[2]++;
            return 1; // singular
        }
    } else {
        return 0;
    }
}

int prune_singular_parent_tracked_longmod_nucifer_stats(unsigned int *g, int n, int maxn) {
    num_graphs_generated[n]++;
    int result = prune_singular_parent_tracked_longmod_nucifer(g,  n, maxn);
    
    if(result == 0)
        num_graphs_accepted[n]++;
    
    return result;
}

int main(int argc, char **argv) {
    if (MAXN > 32) {
        fprintf(stderr, "Currently only supports only up to 32 vertices\n");
        return 3;
    }
    
    //TODO: set max order (cf. nutgen.c)
    
    fprintf(stderr, "Info: Using prune_singular_parent_tracked_longmod_nucifer_stats and preprune\n");
    //int exit_code = pgeng(argc, argv, prune_singular_parent_tracked_longmod_nucifer_stats, 0);
    int exit_code = pgeng(argc, argv, prune_singular_parent_tracked_longmod_nucifer_stats, preprune_singular_parent_tracked_longmod_nucifer);
    
    fprintf(stderr, "Printing statistics:\n");
    
    int i;
    for(i = 1; i <= MAXN; i++)
        if(num_graphs_generated[i] > 0)
            fprintf(stderr, "Nv=%d, num graphs generated: %llu, accepted: %llu (%.1f%%)\n", i, num_graphs_generated[i], num_graphs_accepted[i], (double) num_graphs_accepted[i] / num_graphs_generated[i] * 100);
    
    unsigned long long int total_counter = counter[0] + counter[1] + counter[2];
    fprintf(stderr, "Pseudo reusable: %llu (%.1f%%)\n", counter[0], (double) counter[0] / total_counter * 100);
    //fprintf(stderr, "Pseudo nonreusable: %llu (%.1f%%)\n", counter[1], (double) counter[1] / total_counter * 100);
    fprintf(stderr, "Pseudo other: %llu (%.1f%%)\n", counter[2], (double) counter[2] / total_counter * 100);
    
    fprintf(stderr, "Times can reject nucifer G-v: %llu (%.1f%%) \n", can_reject_nucifer_G_minus_last, (double) can_reject_nucifer_G_minus_last / (can_reject_nucifer_G_minus_last + cannot_reject_nucifer_G_minus_last) * 100);
    
    return exit_code;
}
