/* nutfilter.c
 * 
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>

#include "graph_config.h"
#include "main.h"
#include "mp.h"
#include "graphutils.h"
#include "pseudo.h"
#include "longmod.h"
//#include "modulo.h"

static unsigned long long int times_pseudo_test_sufficient = 0;
static unsigned long long int times_pseudo_test_not_sufficient = 0;

// version without pseudorank test
int is_a_nut_slow(ADJ_LIST_PTR graph, unsigned int *degrees, int nv) {
    MP_MATRIX matrix;
    to_mp_matrix(graph, degrees, nv, matrix);
    int rank = reduce_mp_matrix(nv, matrix);
    
    frequency_rank[rank]++;
    
    if (rank == nv - 1) {
        //print_mp_matrix(stderr, nv, matrix);
        int result = check_zeroless_kernel_vector_mp_matrix(nv, matrix);
        clear_mp_matrix(nv, matrix);
        return result;
    } else {
        clear_mp_matrix(nv, matrix);
        return 0;
    }
}

int is_a_nut(ADJ_LIST_PTR graph, unsigned int *degrees, int nv) {

    // first filter out those graphs that have full 'pseudo' rank
    INT_ADJ_MATRIX matrix;
    to_int_adj_matrix(graph, degrees, nv, matrix);
    
    //if (pseudo_reduce_matrix(nv, matrix) == nv) { //Org
    //if(inverse_modulo_matrix(nv, matrix)) { //Other, don't forget init_modulo() if using this! Is not really faster
//    if(inverse_longmod_matrix(nv, matrix)) { //New

    //Is faster than inverse_longmod_matrix(nv, matrix), since that method saves and diagonalises the matrix!
    int rank = longmod_reduce_matrix(nv, matrix); 
    
    if(rank == nv) {
        times_pseudo_test_sufficient++;
        frequency_rank[nv]++;
        return 0; //Full rank certainly correct, so this is definitely not a nut
    } else {
        times_pseudo_test_not_sufficient++;
        
        // no luck - use multi precision version
        return is_a_nut_slow(graph, degrees, nv);
    }
}

//Warning: this method is not necessarily exact for nv > 22!
//But should be about as exact as analytical methods
int is_a_nut_longmod(ADJ_LIST_PTR graph, unsigned int *degrees, int nv) {
    INT_ADJ_MATRIX matrix;
    to_int_adj_matrix(graph, degrees, nv, matrix);
    
    int rank = longmod_reduce_matrix(nv, matrix);
    
    if(rank == nv)
        return 0; //Full rank certainly correct, so this is definitely not a nut
    else
        return 1; //Could be a nut
    
    //The method below is not guaranteed to be correct if nv > 22
/*
    frequency_rank[rank]++;
    
    if (rank == nv - 1) {
        int result = check_zeroless_kernel_vector_longmod(nv, matrix);
        return result;
    } else {
        return 0;
    }    
*/

}

void print_rank(ADJ_LIST_PTR graph, unsigned int *degrees, int nv) {
    MP_MATRIX matrix;
    to_mp_matrix(graph, degrees, nv, matrix);
    int rank = reduce_mp_matrix(nv, matrix);

    clear_mp_matrix(nv, matrix);
    fprintf(stdout, "%d\n", rank);
}

void print_pseudo_rank(ADJ_LIST_PTR graph, unsigned int *degrees, int nv) {
    INT_ADJ_MATRIX matrix;
    to_int_adj_matrix(graph, degrees, nv, matrix);
    int rank = pseudo_reduce_matrix(nv, matrix);
    fprintf(stdout, "%d\n", rank);
}

int main(int argc, char **argv) {
    
    int cnt = 0;
    int c;
    int command = 'n';
    while ((c = getopt(argc, argv, "prnsl")) != -1) {
        switch (c) {
            case 'p':
            case 'r':
            case 's':
            case 'l':
            case 'n':
                if (cnt == 0) {
                    command = c;
                    cnt++;
                    break;
                }
                // false through in case of error
            default:
                fprintf(stderr, "Usage: %s [-r|-p|-n|-s|-l]\n", argv[0]);
                exit(1);
        }
    }
    switch (command) {
        case 'n':
            main_filter_adj_list(&is_a_nut);
            break;
        case 'l':
            fprintf(stderr, "Warning: the outputted graphs are not necessarily nuts!\n");
            main_filter_adj_list(&is_a_nut_longmod);
            break;            
        case 'p':
            main_process_adj_list(&print_pseudo_rank);
            break;
        case 's':
            main_filter_adj_list(&is_a_nut_slow);
            break;
        default:
            main_process_adj_list(&print_rank);
            break;
    }
    
    fprintf(stderr, "Times pseudo test sufficient: %llu (%.2f%%)\n", times_pseudo_test_sufficient,
                (double) 100 * times_pseudo_test_sufficient / (times_pseudo_test_sufficient + times_pseudo_test_not_sufficient));
}

