/* positionzero.c
 * 
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>

#include "graph_config.h"
#include "main.h"
#include "graphutils.h"
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>


#define EPSILON 0.00001

static unsigned long long int position_NBO[MAXN+2] = {0};


#define MAX_RATIO 10000

//Factor by which the ratio is multiplied
//#define RATIO_MULTIPLICATION_FACTOR 2.0
//#define RATIO_MULTIPLICATION_FACTOR 6.0
//#define RATIO_MULTIPLICATION_FACTOR 30.0
#define RATIO_MULTIPLICATION_FACTOR 120.0

static unsigned long long int ratio_largest_smallest[MAX_RATIO+1] = {0};

static int nv_global = -1;

//Compute the ratio of the largest / smallest entry of the zero eigenvector
void determine_ratio_largest_smallest_eigenvector_entry(gsl_vector_view evec_i, int nv) {
    double smallest_abs_value = DBL_MAX;
    double largest_abs_value = -1;
    //fprintf(stderr, "smallest: %f, largest: %f\n", smallest_abs_value, largest_abs_value);

    //fprintf(stderr, "vec:\n");
    int k;
    for(k = 0; k < nv; k++) {
        //fprintf(stderr, "%f\n", gsl_vector_get(&evec_i.vector, k));
        double d_org = gsl_vector_get(&evec_i.vector, k);
        double d_abs = fabs(d_org);

        //fprintf(stderr, "d_org: %f, d_abs: %f\n", d_org, d_abs);

        if(d_abs < smallest_abs_value)
            smallest_abs_value = d_abs;

        if(d_abs > largest_abs_value)
            largest_abs_value = d_abs;
    }

    if(smallest_abs_value == DBL_MAX || largest_abs_value == -1) {
        fprintf(stderr, "Error: wrong smallest and largest eigenvals!\n");
        exit(1);
    }

    double ratio = RATIO_MULTIPLICATION_FACTOR * largest_abs_value / smallest_abs_value;
    //fprintf(stderr, "%f / %f = %f\n", largest_abs_value, smallest_abs_value, ratio);

    
    //+0.5 since compiler will truncate it when casting
    int ratio_int = (int) (ratio + 0.5);
    //int ratio_int = round(ratio); //Also possible, but then need to compile with -lm
    double difference = ratio - ratio_int;
    
    //fprintf(stderr, "difference: %f\n", difference);
    if(difference > EPSILON || difference < -EPSILON) {
        fprintf(stderr, "Error: is ratio not an int: %d vs %f (factor: %f)\n", ratio_int, ratio, RATIO_MULTIPLICATION_FACTOR);
        fprintf(stderr, "Largest entry: %f, smallest entry: %f\n", largest_abs_value, smallest_abs_value);
        fprintf(stderr, "%f * %f / %f = %f\n", RATIO_MULTIPLICATION_FACTOR, largest_abs_value, smallest_abs_value, ratio);
        fprintf(stderr, "ratio int: %d\n", ratio_int);
        fprintf(stderr, "difference: %f\n", difference);        
        exit(1);
    }

    if(ratio_int > MAX_RATIO) {
        fprintf(stderr, "Error: ratio is too big, increase MAX_RATIO!\n");
        exit(1);
    }    
    
    ratio_largest_smallest[ratio_int]++;
    
}


int determine_position_NBO(ADJ_LIST_PTR graph, unsigned int *degrees, int nv) {
    
    if(nv_global == -1)
        nv_global = nv;
    else if(nv_global != nv) {
        fprintf(stderr, "Error: different nv!\n");
        exit(1);
    }
        
    //print_graph(stderr, graph, degrees, nv);

    INT_ADJ_MATRIX matrix;
    to_int_adj_matrix(graph, degrees, nv, matrix);

    double data[nv*nv];
    int counter = 0;
    for(int i = 0; i < nv; i++)
        for(int j = 0; j < nv; j++)
            data[counter++] = matrix[i][j];

    gsl_matrix_view m = gsl_matrix_view_array(data, nv, nv);
    //gsl_matrix_view m = gsl_matrix_view_array (matrix, nv, nv);

    gsl_vector *eval = gsl_vector_alloc(nv);
    gsl_matrix *evec = gsl_matrix_alloc(nv, nv);

    gsl_eigen_symmv_workspace * w =
            gsl_eigen_symmv_alloc(nv);

    gsl_eigen_symmv(&m.matrix, eval, evec, w);

    gsl_eigen_symmv_free(w);

    
    //Actually we don't need to sort, only to count the number of eigenvalues > 0
    //But sorting is not a bottleneck...

    //gsl_eigen_symmv_sort(eval, evec, GSL_EIGEN_SORT_VAL_ASC);
    //PWF: Definitely the practice is to count down from the largest positive eigenvalue
    gsl_eigen_symmv_sort(eval, evec, GSL_EIGEN_SORT_VAL_DESC);
    
    //Warning: beware of rounding errors!
    //eigenvalue = 5.24834e-16    
    
    int index_zero = -1;
    
    int i;
    for(i = 0; i < nv; i++) {
        double eval_i = gsl_vector_get(eval, i);
        //gsl_vector_view evec_i = gsl_matrix_column(evec, i);

        //fprintf(stderr, "eigenvalue = %g\n", eval_i);
        //fprintf(stderr, "eigenvector = \n");
        //gsl_vector_fprintf(stderr, &evec_i.vector, "%g");
        
        if(-EPSILON < eval_i && eval_i < EPSILON) {
            if(index_zero != -1) {
                fprintf(stderr, "Error: multiple zero eigenvalues!\n");
                exit(1);
            }
            //index_zero = i;
            index_zero = i + 1; //Count from 1
            
            //fprintf(stderr, "Zero eigenvector:\n");
            //gsl_vector_view evec_i = gsl_matrix_column(evec, i);
            //gsl_vector_fprintf(stderr, &evec_i.vector, "%g");
            
            gsl_vector_view evec_i = gsl_matrix_column(evec, i);
            determine_ratio_largest_smallest_eigenvector_entry(evec_i, nv);
            
        }
    }
    
    if(index_zero == -1) {
        fprintf(stderr, "Error: no zero eigenvalue found!\n");
        exit(1);
    }
    
    position_NBO[index_zero]++;
    
    gsl_vector_free(eval);
    gsl_matrix_free(evec);
    
}

int main(int argc, char **argv) {
    
    main_filter_adj_list(&determine_position_NBO);
    
    unsigned long long int total_num_graphs = 0;
    fprintf(stderr, "Frequency table position NBO ceil(n/2) + k, (i.e. %d + k)\n", (nv_global + 1) / 2);
    fprintf(stderr, "Warning: sorting eigenvalues in descending order and start counting from one!\n");
    int i;
    for(i = 0; i <= MAXN+1; i++)
        if(position_NBO[i] > 0) {
            //fprintf(stderr, "position NBO %d: %llu\n", i, position_NBO[i]);
            fprintf(stderr, "position NBO k=%d: %llu\n", i - (nv_global + 1) / 2, position_NBO[i]);
            total_num_graphs += position_NBO[i];
        }
    
    fprintf(stderr, "Total num graphs: %llu\n", total_num_graphs);
    
    
    fprintf(stderr, "Info: RATIO_MULTIPLICATION_FACTOR is %f\n", RATIO_MULTIPLICATION_FACTOR);
    
    total_num_graphs = 0;
    for(i = 0; i <= MAX_RATIO; i++)
        if(ratio_largest_smallest[i] > 0) {
            //fprintf(stderr, "position NBO %d: %llu\n", i, position_NBO[i]);
            fprintf(stderr, "ratio largest/smallest=%f: %llu\n", (double) (i / RATIO_MULTIPLICATION_FACTOR) , ratio_largest_smallest[i]);
            total_num_graphs += ratio_largest_smallest[i];
        }    
    fprintf(stderr, "Total num graphs: %llu\n", total_num_graphs);
    
}

