/* pseudo.c
 *
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file
 * Computes 'pseudo'-ranks of adjacency matrices. Ignores overflows in
 * all computations and hence essentially computes the Z_{2^n}-rank where
 * n is the word size. The pseudo-rank is never larger than the real rank, so
 * if the pseudo-rank is full, the matrix is certain to be nonsingular.
 */

#include <memory.h>
#include <stdio.h>
#include "graph_config.h"

static INT_ADJ_MATRIX inverse;

/*
 * Rescales a row
 */
static
void row_scale(unsigned int *row, int nv, int factor) {
    for (int i = 0; i < nv; i++) {
        row[i] *= factor;
    }
}

/*
 * Reduces destination row
 */
static
void row_reduce(int nv, INT_ADJ_MATRIX_PTR matrix, int dest, int dest_factor, int src, int src_factor) {
    //fprintf (stderr,"row_reduce(%d,%p,%d,%d,%d,%d)\n", nv, matrix, dest, dest_factor, src, src_factor);
    if (src_factor != 0) {
        for (int i = 0; i < nv; i++) {
            matrix[dest][i] = matrix[dest][i] * dest_factor - matrix[src][i] * src_factor;
        }
    }
}

/**
 * Interchanges two rows
 */
static
void row_interchange(int nv, INT_ADJ_MATRIX_PTR matrix, int r1, int r2) {
    for (int i = 0; i < nv; i++) {
        unsigned int tmp = matrix[r1][i];
        matrix[r1][i] = matrix[r2][i];
        matrix[r2][i] = tmp;
    }
}

// for debugging
static
void print_adj_matrix(FILE *f, int nv, INT_ADJ_MATRIX_PTR matrix) {
    for (int i = 0; i < nv; i++) {
        for (int j = 0; j < nv; j++)
            fprintf(f, "%d ", matrix[i][j]);
        fprintf(f, "\n");
    }
}

/**
 * Performs a reduction of the given matrix and returns the rank
 */
int pseudo_reduce_matrix(int nv, INT_ADJ_MATRIX_PTR matrix) {
    int row = 0;
    for (int column = 0; column < nv; column++) {
        // find pivot
        int k = row;
        while (k < nv && matrix[k][column] == 0)
            k++;
        if (k < nv) {
            if (k > row) {
                row_interchange(nv, matrix, k, row);
            }
            for (int dest = row + 1; dest < nv; dest++) {
                row_reduce(nv, matrix, dest, matrix[row][column], row, matrix[dest][column]);
            }
            row++;
        }
    }
    return row;
}

/**
 * Compute the inverse of the given matrix and stores it locally for use in other methods
 * @param[in out] matrix matrix to be inverted
 * @return 1 when matrix could be diagonalized, -1 when it is nonsingular but determinant is 0 modulo WORD_SIZE, 0 otherwise
 */
int near_inverse_pseudo_matrix(int nv, INT_ADJ_MATRIX_PTR matrix) {

    // start with the identity matrix
    memset(inverse, 0, MAXN * MAXN * sizeof(int));
    for (int i = 0; i < nv; i++) {
        inverse[i][i] = 1;
    }

    // do the reduction. Apply all elementary operations to both matrix and inverse
    for (int column = 0; column < nv; column++) {
        // find pivot
        int k = column;
        while (k < nv && matrix[k][column] == 0)
            k++;
        if (k < nv) {
            if (k > column) {
                row_interchange(nv, matrix, k, column);
                row_interchange(nv, inverse, k, column);
            }
            for (int dest = 0; dest < nv; dest++) {
                if (dest != column) {
                    int src_factor = matrix[dest][column];
                    row_reduce(nv, matrix, dest, matrix[column][column], column, src_factor);
                    // important! do not inline src_factor, for matrix[dest][column] becomes 0 after the above call
                    row_reduce(nv, inverse, dest, matrix[column][column], column, src_factor);
                }
            }
        } else {
            return 0; // singular
        }

    }

    // rescale rows so that all diagonal elements are equal
    for (int row = 0; row < nv; row++) {
        int factor = 1;
        for (int i = 0; i < nv; i++) {
            if (i != row) {
                factor *= matrix[i][i];
            }
        }
        if (factor == 0) {
            return -1; // diagonal contains a zero divisor :-(
        }

        row_scale(inverse[row], nv, factor);
    }


    return 1; // true
}

/**
 * Computes bMb^T for a bit vector b and a matrix M.
 */
unsigned int pseudo_norm_bit_vector(unsigned int b, int nv) {
    unsigned int norm = 0;
    for (int i = 0, bitsi = b; i < nv; i++, bitsi <<= 1) {
        if ((bitsi & 0x80000000) != 0) {
            for (int j = 0, bitsj = b; j < nv; j++, bitsj <<= 1) {
                if ((bitsj & 0x80000000) != 0) {
                    norm += inverse[i][j];
                }
            }
        }
    }
    return norm;
}


